import base64
from urllib.parse import urlparse

from net_file.common import NetFile
from net_file.local_file import supported_schemas as local_schemas
from net_file.local_file import open_local_file
from net_file.http_file import supported_schemas as http_schemas
from net_file.http_file import open_http_url


_SCHEMA_HANDLERS = {
    local_schemas(): open_local_file,
    http_schemas(): open_http_url,
}


def open_url(
        url: str,
        start_bytes: int = None,
        length: int = None,
) -> NetFile:
    parsed_url = urlparse(url)
    scheme = parsed_url.scheme.lower()
    for schemas, handler in _SCHEMA_HANDLERS.items():
        if scheme in schemas:
            return handler(
                url=url,
                start_bytes=start_bytes,
                length=length,
            )
    raise ValueError(f'Unknown schema {parsed_url.scheme}')



from os import path

from net_file.net_file import open_url


def test_read_local_file():
    readme_file_path = path.normpath(path.join(path.abspath(path.dirname(__file__)), '..', 'examples', 'samples', 'test_file.txt'))

    parts = []
    with open_url(
            url=f'file://{readme_file_path}',
            start_bytes=20,
            length=30,
    ) as open_file:
        content = open_file.read(7)
        while content:
            parts.append(content)
            content = open_file.read(7)
    assert parts == [b'2222222', b'2223333', b'3333334', b'4444444', b'44']



class NetFile:
    def __init__(self):
        pass

    def __enter__(self):
        self.open()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def open(self):
        raise NotImplementedError()

    def close(self):
        raise NotImplementedError()

    def read(self, chunk_bytes: int = None) -> bytes:
        raise NotImplementedError()


class NetFileError(Exception):
    pass


class InvalidFileRequestError(NetFileError):
    pass


class FileUnavailableError(NetFileError):
    pass


class FileMissingError(InvalidFileRequestError):
    pass


class InvalidFileRangeError(InvalidFileRequestError):
    pass



from pathlib import Path
from typing import Tuple

from net_file.common import FileMissingError
from net_file.common import FileUnavailableError
from net_file.common import InvalidFileRangeError
from net_file.common import InvalidFileRequestError
from net_file.common import NetFile


class LocalFile(NetFile):
    def __init__(
            self,
            path: Path,
            start_bytes: int = None,
            length: int = None,
    ):
        super().__init__()
        if start_bytes < 0 or length < 0:
            raise InvalidFileRangeError()

        self._path = path
        self._start_bytes = start_bytes or 0
        self._length = length
        self._raw_file = None
        self._remaining_bytes = 0

    def read(self, chunk_bytes: int = None) -> bytes:
        assert chunk_bytes is None or chunk_bytes > 0

        if self.closed:
            raise InvalidFileRequestError('File is closed')

        content = b''
        if self._remaining_bytes:
            try:
                if chunk_bytes is not None:
                    content = self._raw_file.read(min(self._remaining_bytes, chunk_bytes))
                else:
                    content = self._raw_file.read(self._remaining_bytes)
            except IOError as exc:
                raise FileUnavailableError(f'File become unavailable ({exc})')

        self._remaining_bytes -= len(content)
        return content

    @property
    def closed(self) -> bool:
        return not self._raw_file or self._raw_file.closed

    def open(self) -> None:
        if self._raw_file and not self._raw_file.closed:
            return

        try:
            self._raw_file = self._path.open(mode='rb')
        except FileNotFoundError:
            raise FileMissingError(f'File {self._path} is missing')
        except IOError as exc:
            raise InvalidFileRequestError(f'Failed to open file ({exc})')

        file_size = self._path.stat().st_size
        if self._start_bytes or self._length:
            length = (self._length or file_size) - self._start_bytes
            self._remaining_bytes = self._length
            if self._start_bytes + length > file_size:
                raise InvalidFileRangeError(f'Invalid file length/start offset specified ({self._length}/{self._start_bytes}). File length: {file_size}')

        if self._start_bytes:
            try:
                self._raw_file.seek(self._start_bytes)
            except IOError:
                raise InvalidFileRangeError(f'Failed to seek file to {self._start_bytes} position')

    def close(self) -> None:
        if not self._raw_file or self._raw_file.closed:
            return

        try:
            self._raw_file.close()
        except IOError:
            pass
        finally:
            self._raw_file = None


def supported_schemas() -> Tuple[str, ...]:
    return 'file',


def open_local_file(
        url: str,
        start_bytes: int = None,
        length: int = None,
) -> NetFile:
    assert url.startswith('file:///')
    return LocalFile(
        path=Path(url[7:]),
        start_bytes=start_bytes,
        length=length,
    )



from os import path

from setuptools import find_packages
from setuptools import setup

readme_file_path = path.join(path.abspath(path.dirname(__file__)), 'README.md')
with open(readme_file_path, encoding='utf-8') as readme_file:
    long_description = readme_file.read()

setup(
    name='net-file',
    version='0.1.0',
    url='https://gitlab.com/kraevs/net-file',
    project_urls={
        'Code': 'https://gitlab.com/kraevs/net-file',
        'Issue tracker': 'https://gitlab.com/kraevs/net-file/issues'
    },
    description='Ranged read file access abstraction.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='Stanislav Kraev',
    author_email='stanislav.kraev@gmail.com',
    license='MIT',
    classifiers=[
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8'
    ],
    packages=find_packages(),
    platforms='any',
    python_requires='>=3.6',
    install_requires=[
        'requests==2.24.0',
    ]
)



from typing import Tuple

import requests

from net_file.common import FileUnavailableError
from net_file.common import InvalidFileRangeError
from net_file.common import InvalidFileRequestError
from net_file.common import NetFile


class HttpFile(NetFile):
    def __init__(
            self,
            url: str,
            start_bytes: int = None,
            length: int = None,
    ):
        super().__init__()
        if start_bytes < 0 or length < 0:
            raise InvalidFileRangeError()

        self._url = url
        self._start_bytes = start_bytes or 0
        self._length = length
        self._remaining_bytes = 0
        self._response = None
        self._response_generator = None

    def open(self) -> None:
        if self._response is not None:
            return

        response = requests.head(url=self._url, headers={'Accept-Encoding': 'identity'})
        try:
            response.raise_for_status()
        except requests.HTTPError as exc:
            raise InvalidFileRequestError(str(exc))

        file_size = int(response.headers['Content-Length'])

        headers = {}
        if self._start_bytes or self._length:
            length = (self._length or file_size) - self._start_bytes
            self._remaining_bytes = self._length
            if self._start_bytes + length > file_size:
                raise InvalidFileRangeError(f'Invalid file length/start offset specified ({self._length}/{self._start_bytes}). File length: {file_size}')
            accept_ranges = response.headers.get('Accept-Ranges')
            if accept_ranges != 'bytes':
                raise InvalidFileRangeError(f"Server doesn't support ranges")

            range_from = self._start_bytes or 0
            range_to = (self._length + range_from - 1) if self._length else (file_size - 1)
            headers['Range'] = f'bytes={range_from}-{range_to}'

        self._response = requests.get(url=self._url, stream=True, headers=headers)
        try:
            self._response.raise_for_status()
        except requests.HTTPError as exc:
            raise InvalidFileRequestError(str(exc))

    def close(self) -> None:
        if not self._response:
            return
        self._response.close()
        self._response = None

    def read(self, chunk_bytes: int = None) -> bytes:
        if not self._response:
            raise InvalidFileRequestError('File is closed')

        def _generator():
            chunk_size = None if chunk_bytes is None else (chunk_bytes or 1024)
            try:
                for chunk in self._response.iter_content(chunk_size=chunk_size):
                    yield chunk
            except requests.RequestException:
                raise FileUnavailableError()

        if not self._response_generator:
            self._response_generator = _generator()

        try:
            return next(self._response_generator)
        except StopIteration:
            self._response_generator = None
            self._response.close()
            self._response = None


def supported_schemas() -> Tuple[str, ...]:
    return 'http', 'https'


def open_http_url(
        url: str,
        start_bytes: int = None,
        length: int = None,
) -> NetFile:
    return HttpFile(url=url, start_bytes=start_bytes, length=length)



import base64

from net_file.net_file import open_url


def tutorial():
    with open_url(
            url='http://10.16.10.131:8888/storage/prepared_datasets/27.12_bord_real/video/02d87f31-ab38-43ae-9d28-bb256c725192.mp4',
            start_bytes=10000,
            length=20000,
    ) as open_file:
        content = open_file.read()
        while content:
            print(len(content), base64.b64encode(content))
            content = open_file.read()

    with open_url(
            url='file:///opt/project/examples/samples/test_file.txt',
            start_bytes=20,
            length=30,
    ) as open_file:
        content = open_file.read(7)
        while content:
            print(content, end='')
            content = open_file.read(7)
        print()

    open_file = open_url(
            url='file:///opt/project/examples/samples/test_file.txt',
            start_bytes=30,
            length=20,
    )
    open_file.open()
    content = open_file.read()
    print(content)
    open_file.close()
    open_file.open()
    content = open_file.read(10000)
    print(content)


if __name__ == "__main__":
    tutorial()
